/******************************************************************************
 *
 * BioHDF - HDF5-based data model and tools for storage and analysis of 
 *          Next Gen Sequencing reads and alignments. 
 *
 * Copyright (c) 2009 Geospiza, Inc.
 *
 * This program can be redistributed and/or modified under the terms of the 
 * Version 2 of the GNU General Public License as published by the Free 
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 *****************************************************************************/


/******************************************************************************
 *
 * BioHDF library functions. 
 *
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <time.h>

#include "hdf5.h"

#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

int _create_sequence_dataset(hid_t gid, int seq_length);

int _create_seqid_dataset(hid_t gid, int seqid_length);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

/* TODO - parse multi-line fasta properly */
int H5BIOparse_fastA(hid_t gid_seqs, int max_seq_length, int is_verbose) {

  int seq_length = -1; /* parsed from first sequence */
  hid_t did_seqid;
  hid_t did_seq; /* created once length known */

  /* create buffer for sequence and seqid data */
  int seq_buf_size = 1024;
  char* seq_buf = NULL;
  int seqid_buf_size = 1024;
  char* seqid_buf = (char*)calloc(seqid_buf_size * SEQID_LENGTH, sizeof(char));

  /* create the seqid dataset */
  did_seqid = _create_seqid_dataset(gid_seqs, SEQID_LENGTH);

  /* parse the csfasta file */
  char line[1024]; /* TODO - parameterize max line length */
  char *the_seqid = (char *)calloc(SEQID_LENGTH, sizeof(char));
  char *seqid = (char *)calloc(SEQID_LENGTH, sizeof(char));
  char seq[SEQ_MAX_LENGTH];
  int seq_num = 0;

  clock_t start_clock, end_clock;
  if (is_verbose) { start_clock = clock(); }

  strcpy(seqid, "");
  while (fgetline(stdin, line, sizeof(line)) > 0) {

    /* skip comments */
    if (line[0] == '#') {
      continue;
    }

    /* parse id */
    if (sscanf(line, ">%s", the_seqid) == 1) {

      /* append previous sequence to dataset */
      if (strlen(seqid) > 0) {

	/* append the seqid to dataset, via a buffer */
	int seqid_buf_loc = seq_num % seqid_buf_size;
	memcpy(seqid_buf + seqid_buf_loc * SEQID_LENGTH, seqid, SEQID_LENGTH);
	if (seqid_buf_loc == seqid_buf_size - 1) {
	  H5UTLwrite_2D_char_buf(did_seqid, seq_num + 1, seqid_buf, seqid_buf_size, SEQID_LENGTH);
	}

	/* append the sequence to dataset, via a buffer */
	int buf_loc = seq_num % seq_buf_size;
	memcpy(seq_buf + buf_loc * seq_length, seq, seq_length);
	if (buf_loc == seq_buf_size - 1) {
	  H5UTLwrite_2D_char_buf(did_seq, seq_num + 1, seq_buf, seq_buf_size, seq_length);
	}

	++seq_num;

	if (is_verbose) {
	  if (seq_num % 1000000 == 0) {
	    end_clock = clock();
	    double elapsed = ((double)(end_clock - start_clock)) / CLOCKS_PER_SEC;
	    start_clock = end_clock;
	    fprintf(stderr, "Processed %d sequences in %g seconds\n", seq_num, elapsed);
	  }
	}

      }
      memset(seqid, 0, SEQID_LENGTH);
      strcpy(seqid, the_seqid);
    } 
    else {
      memset(seq, 0, SEQ_MAX_LENGTH);
      sscanf(line, "%s", seq);

      if (seq_length < 0) {
	if (max_seq_length < 0) {
	  seq_length = strlen(seq);
	} else {
	  seq_length = max_seq_length;
	}

	/* create the sequence dataset and write buffer */
        did_seq = _create_sequence_dataset(gid_seqs, seq_length);
	seq_buf = (char*)calloc(seq_buf_size * seq_length, sizeof(char));
      }

    }
  }

  /* append previous sequence to dataset */
  if (strlen(seqid) > 0) {
    
    /* append the seqid to dataset, via a buffer */
    int seqid_buf_loc = seq_num % seqid_buf_size;
    memcpy(seqid_buf + seqid_buf_loc * SEQID_LENGTH, seqid, SEQID_LENGTH);
    H5UTLwrite_2D_char_buf(did_seqid, seq_num + 1, seqid_buf, seqid_buf_loc + 1, SEQID_LENGTH);
      
    /* append the sequence to dataset, via a buffer */
    int buf_loc = seq_num % seq_buf_size;
    memcpy(seq_buf + buf_loc * seq_length, seq, seq_length);
    H5UTLwrite_2D_char_buf(did_seq, seq_num + 1, seq_buf, buf_loc + 1, seq_length);
      
    ++seq_num;
  }
  if (did_seq) { H5Dclose(did_seq); }

  if (is_verbose) { fprintf(stderr, "Processed %d sequences\n", seq_num); }

  return 0;
}


int H5BIOparse_fastQ(hid_t gid_seqs, int max_seq_length, int is_verbose) {

  int seq_length = -1; /* parsed from first sequence */
  hid_t did_seqid;
  hid_t did_seq; /* created once length known */

  /* create buffer for sequence and seqid data */
  int seq_buf_size = 1024;
  char* seq_buf = NULL;
  int seqid_buf_size = 1024;
  char* seqid_buf = (char*)calloc(seqid_buf_size * SEQID_LENGTH, sizeof(char));

  /* create the seqid dataset */
  did_seqid = _create_seqid_dataset(gid_seqs, SEQID_LENGTH);

  /* parse the fastq file */
  char line[1024]; /* TODO - parameterize max line length */
  char *seqid = (char *)calloc(SEQID_LENGTH, sizeof(char));
  char seq[80];
  int seq_num = 0;

  clock_t start_clock, end_clock;
  if (is_verbose) { start_clock = clock(); }

  strcpy(seqid, "");
  while (fgetline(stdin, line, sizeof(line)) > 0) {

    /* parse id */
    if (sscanf(line, "@%s", seqid) == 1) {

      /* parse the sequence line */
      if (fgetline(stdin, line, sizeof(line)) <= 0) {
	fprintf(stderr, "bad fastq format\n");
	return -1;
      }
      sscanf(line, "%s", seq);

      if (seq_length < 0) {
	if (max_seq_length < 0) {
	  seq_length = strlen(seq);
	} else {
	  seq_length = max_seq_length;
	}

	/* create the sequence dataset and write buffer */
        did_seq = _create_sequence_dataset(gid_seqs, seq_length);
	seq_buf = (char*)calloc(seq_buf_size * seq_length, sizeof(char));
      }

      /* append the seqid to dataset, via a buffer */
      int seqid_buf_loc = seq_num % seqid_buf_size;
      memcpy(seqid_buf + seqid_buf_loc * SEQID_LENGTH, seqid, SEQID_LENGTH);
      if (seqid_buf_loc == seqid_buf_size - 1) {
	H5UTLwrite_2D_char_buf(did_seqid, seq_num + 1, seqid_buf, seqid_buf_size, SEQID_LENGTH);
      }
      
      /* append the sequence to dataset, via a buffer */
      int buf_loc = seq_num % seq_buf_size;
      memcpy(seq_buf + buf_loc * seq_length, seq, seq_length);
      if (buf_loc == seq_buf_size - 1) {
	H5UTLwrite_2D_char_buf(did_seq, seq_num + 1, seq_buf, seq_buf_size, seq_length);
      }
      
      ++seq_num;
      
      if (is_verbose) {
	if (seq_num % 1000000 == 0) {
	  end_clock = clock();
	  double elapsed = ((double)(end_clock - start_clock)) / CLOCKS_PER_SEC;
	  start_clock = end_clock;
	  fprintf(stderr, "Processed %d sequences in %g seconds\n", seq_num, elapsed);
	}
      }

      /* parse the 2nd sequence ID line */
      if (fgetline(stdin, line, sizeof(line)) <= 0) {
	fprintf(stderr, "bad fastq format\n");
	return -1;
      }

      /* parse the quality line */
      if (fgetline(stdin, line, sizeof(line)) <= 0) {
	fprintf(stderr, "bad fastq format\n");
	return -1;
      }
    } 
    else {
      fprintf(stderr, "bad fastq format\n");
      return -1;
    }
  }

  /* flush the seqid buffer to dataset */
  int seqid_buf_loc = seq_num % seqid_buf_size;
  H5UTLwrite_2D_char_buf(did_seqid, seq_num, seqid_buf, seqid_buf_loc, SEQID_LENGTH);
      
  /* flush the seqid buffer to dataset */
  int buf_loc = seq_num % seq_buf_size;
  H5UTLwrite_2D_char_buf(did_seq, seq_num, seq_buf, buf_loc, seq_length);

  if (did_seq) { H5Dclose(did_seq); }

  if (is_verbose) { fprintf(stderr, "Processed %d sequences\n", seq_num); }

  return 0;
}


/* write alignment hit buffer into the dataset */
void write_hit_buf(hid_t did_aln, int hit_num, const hit_t* hit_buf, int hit_buf_num, hid_t hit_tid) {

  hsize_t dims[1] = { hit_num };
  H5Dextend(did_aln, dims);
  hid_t sid_aln = H5Dget_space(did_aln);
  
  hsize_t slab_start[1] = { hit_num - hit_buf_num };
  hsize_t slab_count[1] = { hit_buf_num };
  H5Sselect_hyperslab(sid_aln, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
  hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
  H5Dwrite(did_aln, hit_tid, mid, sid_aln, H5P_DEFAULT, hit_buf);
  H5Sclose(mid);
}


int compare_refmap(const void *a, const void *b) {
  refmap_t *rma = (refmap_t *)a;
  refmap_t *rmb = (refmap_t *)b;
  int cmp_ref_id = (int)(rma->ref_id - rmb->ref_id);
  if (cmp_ref_id != 0) {
    return cmp_ref_id;
  } else {
    int cmp_position = (int)(rma->position - rmb->position);
    if (cmp_position != 0) {
      return cmp_position;
    } else {
      return (int)(rma->hit_id - rmb->hit_id);
    }
  }
}


/* create the refmap memory datatype */
hid_t H5BIOhit_type() {

  hid_t tid = H5Tcreate(H5T_COMPOUND, sizeof(hit_t));

  H5Tinsert(tid, "read_id",  HOFFSET(hit_t, read_id),  H5T_NATIVE_INT);
  H5Tinsert(tid, "ref_id",   HOFFSET(hit_t, ref_id),   H5T_NATIVE_INT);
  H5Tinsert(tid, "position", HOFFSET(hit_t, position), H5T_NATIVE_INT);
  H5Tinsert(tid, "length",   HOFFSET(hit_t, length),   H5T_NATIVE_INT);
  H5Tinsert(tid, "strand",   HOFFSET(hit_t, strand),   H5T_NATIVE_INT);
  H5Tinsert(tid, "mismatch", HOFFSET(hit_t, mismatch), H5T_NATIVE_INT);

  return tid;
}


/* create the refmap memory datatype */
hid_t H5BIOrefmap_type() {

  hid_t tid = H5Tcreate(H5T_COMPOUND, sizeof(refmap_t));

  H5Tinsert(tid, "ref_id",   HOFFSET(refmap_t, ref_id),   H5T_NATIVE_INT);
  H5Tinsert(tid, "position", HOFFSET(refmap_t, position), H5T_NATIVE_INT);
  H5Tinsert(tid, "length",   HOFFSET(refmap_t, length),   H5T_NATIVE_INT);
  H5Tinsert(tid, "hit_id",   HOFFSET(refmap_t, hit_id),   H5T_NATIVE_INT);

  return tid;
}


/* create the cluster memory datatype */
hid_t H5BIOcluster_type() {

  hid_t tid = H5Tcreate(H5T_COMPOUND, sizeof(cluster_t));

  H5Tinsert(tid, "ref_id",   HOFFSET(cluster_t, ref_id),   H5T_NATIVE_INT);
  H5Tinsert(tid, "beg_pos",  HOFFSET(cluster_t, beg_pos),  H5T_NATIVE_INT);
  H5Tinsert(tid, "end_pos",  HOFFSET(cluster_t, end_pos),  H5T_NATIVE_INT);
  H5Tinsert(tid, "num_read", HOFFSET(cluster_t, num_read), H5T_NATIVE_INT);

  return tid;
}


/* create the reference sequence index memory datatype */
hid_t H5BIOrefidx_type() {

  hid_t tid = H5Tcreate(H5T_COMPOUND, sizeof(refidx_t));

  H5Tinsert(tid, "offset",   HOFFSET(refidx_t, offset),   H5T_NATIVE_LLONG);
  H5Tinsert(tid, "length",   HOFFSET(refidx_t, length),   H5T_NATIVE_INT);

  return tid;
}


int H5BIOreadmask_alignment(hid_t fid, hid_t gid_aln) {

  /* nothing to do if readmask already exists */
  if (H5UTLdataset_exists(gid_aln, READMASK_DATASET)) {
    return 0;
  }

  hid_t did_aln;
  HDFRET(did_aln = H5Dopen(gid_aln, ALIGNMENT_DATASET, H5P_DEFAULT));

  /* reads must be imported before alignment readmask computed */
  if (!H5UTLgroup_exists(fid, SEQUENCES_GROUP)) {
    fprintf(stderr, "Group '%s' must exist before computing alignment readmask\n", SEQUENCES_GROUP);
    return -1;
  }

  /* get fixed sequence length from 2nd dim of 'sequence' dataset extent */
  char hdf_path[PATH_MAX];
  sprintf(hdf_path, "/%s/%s", SEQUENCES_GROUP, SEQUENCE_DATASET);
  hid_t did_seq, sid_seq;
  HDFRET(did_seq = H5Dopen(fid, hdf_path, H5P_DEFAULT));
  HDFRET(sid_seq = H5Dget_space(did_seq));
  int num_seq, seq_len;
  H5UTLget_2D_extent(sid_seq, &num_seq, &seq_len);
  H5Sclose(sid_seq);
  H5Dclose(did_seq);

  /* initialize the readmask dataset */
  char *readmask = (char *)calloc(num_seq, sizeof(char));

  /* iterate over the alignment */
  hid_t sid_aln;
  HDFRET(sid_aln = H5Dget_space(did_aln));
  int hit_num = H5UTLget_1D_extent(sid_aln);

  /* use fixed-size buffer */
  int buf_size = HDF_RW_BUF_SIZE;
  hit_t *hit_buf = (hit_t *)calloc(buf_size, sizeof(hit_t));

  /* create the hit memory datatype */
  hid_t tid_hit;
  HDFRET(tid_hit = H5BIOhit_type());

  /* iterate over all hits */
  int i;
  hit_t *item;
  int buf_offset = - buf_size;
  for (i = 0; i < hit_num; i++) {
    if (i >= buf_offset + buf_size) {
      /* fill buffer */
      buf_offset += buf_size;
      int buf_fill = hit_num - buf_offset;
      if (buf_fill > buf_size) {
	buf_fill = buf_size;
      }
      hsize_t slab_start[1] = { buf_offset };
      hsize_t slab_count[1] = { buf_fill };
      H5Sselect_hyperslab(sid_aln, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
      H5Dread(did_aln, tid_hit, mid, sid_aln, H5P_DEFAULT, hit_buf);
      H5Sclose(mid);
    }
    item = hit_buf + i - buf_offset;

    readmask[item->read_id] = 1;
  }

  /* store the read mask */
  hid_t did_mask;
  HDFRET(did_mask = H5UTLcreate_1D_dataset_size(gid_aln, READMASK_DATASET, H5T_NATIVE_CHAR, num_seq));
  hsize_t dims[1] = { num_seq };
  H5Dextend(did_mask, dims);
  H5Dwrite(did_mask, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, readmask);

  return 0;
}


int H5BIOindex_alignment(hid_t fid, hid_t gid_aln) {

  /* nothing to do if index already exists */
  if (H5UTLdataset_exists(gid_aln, REFMAP_DATASET)) {
    return 0;
  }

  hid_t tid_hit, tid_refmap, did_aln, did_map, sid_aln;

  /* create the hit memory datatype */
  HDFRET(tid_hit = H5BIOhit_type());

  /* create the refmap memory datatype */
  HDFRET(tid_refmap = H5BIOrefmap_type());

  /* open the "alignment" dataset (list of individual read/ref hits) */
  HDFRET(did_aln = H5Dopen(gid_aln, ALIGNMENT_DATASET, H5P_DEFAULT));

  /* create the index dataset 
     (list of hits sorted on numeric ref_id then on hit position) */
  HDFRET(did_map = H5UTLcreate_1D_dataset(gid_aln, REFMAP_DATASET, tid_refmap));

  /* iterate through the alignment dataset, creating index to be sorted */
  HDFRET(sid_aln = H5Dget_space(did_aln));
  int hit_num = H5UTLget_1D_extent(sid_aln);

  /* use fixed-size buffer */
  int buf_size = HDF_RW_BUF_SIZE;
  hit_t *hit_buf = (hit_t *)malloc(sizeof(hit_t) * buf_size);

  /* an index buffer to be sorted later */
  void *refmap_buf = malloc(sizeof(refmap_t) * hit_num);
  refmap_t *refmap;

  /* iterate over all hits */
  int i;
  hit_t *item;
  int buf_offset = - buf_size;
  for (i = 0, refmap = refmap_buf; i < hit_num; i++, refmap++) {
    if (i >= buf_offset + buf_size) {
      /* fill buffer */
      buf_offset += buf_size;
      int buf_fill = hit_num - buf_offset;
      if (buf_fill > buf_size) {
	buf_fill = buf_size;
      }
      hsize_t slab_start[1] = { buf_offset };
      hsize_t slab_count[1] = { buf_fill };
      H5Sselect_hyperslab(sid_aln, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
      H5Dread(did_aln, tid_hit, mid, sid_aln, H5P_DEFAULT, hit_buf);
      H5Sclose(mid);
    }
    item = hit_buf + i - buf_offset;

    refmap->ref_id = item->ref_id;
    refmap->position = item->position;
    refmap->length = item->length;
    refmap->hit_id = i;
  }

  /* sort the refmap buffer, storing to a dataset as an index into hits */
  H5UTLsort_store_1D_dataset(did_map, tid_refmap,
			     refmap_buf, hit_num, sizeof(refmap_t), 
			     compare_refmap);

  return 0;
}


void refmap_item(refmap_t *item, hid_t did, hid_t sid, hid_t tid, hid_t mid, int pp) {

  hsize_t slab_start[1] = { pp };
  hsize_t slab_count[1] = { 1 };
  H5Sselect_hyperslab(sid, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
  H5Dread(did, tid, mid, sid, H5P_DEFAULT, item);
}


int refmap_is_match(refmap_t *item, 
		    int ref_id, int start_pos, int end_pos, int seq_len) {
  if (item->hit_id < 0) { /* signals 'no item available' */
    return 0; 
  } else {
    return (item->ref_id == ref_id &&
	    item->position + 1 <= end_pos &&
	    item->position + item->length >= start_pos);
  }
}


int refmap_is_first(refmap_t *item, refmap_t *prior_item, 
		    int ref_id, int start_pos, int end_pos, int seq_len) {
  /* NOTE - only works because reads are identical length */
  return (!refmap_is_match(prior_item, ref_id, start_pos, end_pos, seq_len) &&
	  refmap_is_match(item, ref_id, start_pos, end_pos, seq_len));
}


int refmap_is_prior(refmap_t *item, refmap_t *prior_item, 
		    int ref_id, int start_pos, int end_pos, int seq_len) {
  return (item->ref_id < ref_id ||
	  (item->ref_id == ref_id && 
	   item->position + item->length < start_pos));
}


int refmap_is_after(refmap_t *item, refmap_t *prior_item, 
		    int ref_id, int start_pos, int end_pos, int seq_len) {
  return (item->ref_id > ref_id ||
	  (item->ref_id == ref_id &&
	   item->position + 1 > end_pos) ||
	  (refmap_is_match(item, ref_id, start_pos, end_pos, seq_len) &&
	   refmap_is_match(prior_item, ref_id, start_pos, end_pos, seq_len)));
}


/* TODO - search using the bounding read length (seq_len) but filter during
   the linear pass, checking for match on the actual read length */

/* binary search to find matching range, then linear pass over range */
void refmap_binary_search(hid_t did, const char *ref_name, 
			  int ref_id, int start_pos, int end_pos,
			  int seq_len, 
			  void (*refmap_action)(const char *, const refmap_t *, int, int, void *),
			  void *data) {
  hsize_t slab_count[1] = { 1 };
  hid_t mid = H5Screate_simple(1, slab_count, slab_count); 

  hid_t tid = H5BIOrefmap_type();

  hid_t sid = H5Dget_space(did);

  refmap_t *prior_item = (refmap_t *)malloc(sizeof(refmap_t));
  refmap_t *item = (refmap_t *)malloc(sizeof(refmap_t));

  int ll = -1;
  int rr = H5UTLget_1D_extent(sid);
  int pp;

  int found = 0;
  while ((pp = (rr - ll) / 2) > 0) {
    pp = ll + pp;
    if (pp > 0) {
      refmap_item(prior_item, did, sid, tid, mid, pp - 1);
    } else {
      prior_item->hit_id = -1;
    }
    refmap_item(item, did, sid, tid, mid, pp);
    if (refmap_is_first(item, prior_item, ref_id, start_pos, end_pos, seq_len)) {
      found = 1;
      break;
    } else if (refmap_is_prior(item, prior_item, ref_id, start_pos, end_pos, seq_len)) {
      ll = pp;
    } else if (refmap_is_after(item, prior_item, ref_id, start_pos, end_pos, seq_len)) {
      rr = pp;
    } else {
      fprintf(stderr, "inconsistent binary search\n");
      exit(1);
    }
  }

  if (found) {
    /* linear pass, printing matching range */
    refmap_linear_search(did, ref_name, ref_id, start_pos, end_pos, pp, seq_len,
			 refmap_action, data);
  }

  /* tidy up */
  H5Tclose(tid);
  H5Sclose(mid);
  H5Sclose(sid);

  free(prior_item);
  free(item);
}


void refmap_linear_search(hid_t did, const char *ref_name, 
			  int ref_id, int start_pos, int end_pos,
			  int start_id, int seq_len, 
			  void (*refmap_action)(const char *, const refmap_t *, int, int, void *),
			  void *data) {

  hid_t sid = H5Dget_space(did);
  int hit_num = H5UTLget_1D_extent(sid);

  hid_t tid = H5BIOrefmap_type();

  int buf_size = HDF_RW_BUF_SIZE;
  refmap_t *map_buf = (refmap_t *)malloc(sizeof(refmap_t) * buf_size);

  /* iterate over all */
  int i;
  refmap_t *item;
  int found = 0;
  int buf_offset = start_id - buf_size;
  for (i = start_id; i < hit_num; i++) {
    if (i >= buf_offset + buf_size) {
      /* fill buffer */
      buf_offset += buf_size;
      int buf_fill = hit_num - buf_offset;
      if (buf_fill > buf_size) {
	buf_fill = buf_size;
      }
      hsize_t slab_start[1] = { buf_offset };
      hsize_t slab_count[1] = { buf_fill };
      H5Sselect_hyperslab(sid, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
      H5Dread(did, tid, mid, sid, H5P_DEFAULT, map_buf);
      H5Sclose(mid);
    }
    item = map_buf + i - buf_offset;
    if (refmap_is_match(item, ref_id, start_pos, end_pos, seq_len)) {
      found = 1;
      (*refmap_action)(ref_name, item, start_pos, end_pos, data);
    } else if (found == 1) {
      break;
    }
  }

  H5Tclose(tid);
  H5Sclose(sid);
}


/* TODO - gff 'source' should be alignment algorithm from HDF attribute */
void refmap_to_gff(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data) {
  printf("%s\tmapreads\tread\t%d\t%d\t.\t.\t.\t%d\n", 
	 ref_name, item->position + 1, item->position + item->length, item->hit_id);
}


void refmap_to_depth(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data) {
  int *depth = (int *)data;
  int width = end_pos - start_pos + 1;
  int i;
  for (i = item->position - start_pos; i < item->position - start_pos + item->length; i++) {
    if (i >= 0 && i < width) {
      ++depth[i];
    }
  }
}

void refmap_to_density(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data) {
  int *density = (int *)data;
  int i;
  for (i = item->position; i < item->position + item->length; i++) {
    if (i < 0 || i >= 100000) {
      fprintf(stderr, "problem\n");
      exit(1);
    }
    ++density[i];
  }
}


/* write refidx entry into the dataset */
void H5BIOwrite_refidx(hid_t did, int num, const refidx_t* buf, int buf_num, hid_t tid) {

  hsize_t dims[1] = { num };
  H5Dextend(did, dims);
  hid_t sid = H5Dget_space(did);
  
  hsize_t slab_start[1] = { num - buf_num };
  hsize_t slab_count[1] = { buf_num };
  H5Sselect_hyperslab(sid, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
  hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
  H5Dwrite(did, tid, mid, sid, H5P_DEFAULT, buf);
  H5Sclose(mid);
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

#define SEQUENCE_RANK 2
int _create_sequence_dataset(hid_t gid, int seq_length) {
  int seq_num = 0;
  hsize_t dims[SEQUENCE_RANK] = { seq_num, seq_length };
  hsize_t maxdims[SEQUENCE_RANK] = { H5S_UNLIMITED, seq_length };
  hid_t sid_seq = H5Screate_simple(SEQUENCE_RANK, dims, maxdims);
  hid_t pid_seq = H5Pcreate(H5P_DATASET_CREATE);
  hsize_t chunk_size[SEQUENCE_RANK] = { 1024, seq_length };
  H5Pset_chunk(pid_seq, SEQUENCE_RANK, chunk_size);
  H5Pset_deflate(pid_seq, HDF_COMPRESSION_LEVEL);
  hid_t did_seq = H5Dcreate(gid, SEQUENCE_DATASET, H5T_NATIVE_CHAR, 
			    sid_seq, H5P_DEFAULT, pid_seq, H5P_DEFAULT);
  H5Pclose(pid_seq);

  return did_seq;
}


#define SEQID_RANK 2
int _create_seqid_dataset(hid_t gid, int seqid_length) {
  int seqid_num = 0;
  hsize_t dims[SEQID_RANK] = { seqid_num, seqid_length };
  hsize_t maxdims[SEQID_RANK] = { H5S_UNLIMITED, seqid_length };
  hid_t sid_seqid = H5Screate_simple(SEQID_RANK, dims, maxdims);
  hid_t pid_seqid = H5Pcreate(H5P_DATASET_CREATE);
  hsize_t chunk_size[SEQID_RANK] = { 1024, seqid_length };
  H5Pset_chunk(pid_seqid, SEQID_RANK, chunk_size);
  H5Pset_deflate(pid_seqid, HDF_COMPRESSION_LEVEL);
  hid_t did_seqid = H5Dcreate(gid, SEQID_DATASET, H5T_NATIVE_CHAR, 
			      sid_seqid, H5P_DEFAULT, pid_seqid, H5P_DEFAULT);
  H5Pclose(pid_seqid);

  return did_seqid;
}


