/******************************************************************************
 *
 * BioHDF - HDF5-based data model and tools for storage and analysis of 
 *          Next Gen Sequencing reads and alignments. 
 *
 * Copyright (c) 2009 Geospiza, Inc.
 *
 * This program can be redistributed and/or modified under the terms of the 
 * Version 2 of the GNU General Public License as published by the Free 
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 *****************************************************************************/


/******************************************************************************
 *
 * BioHDF library header. 
 *
 *****************************************************************************/

#ifndef BIOHDF_H
#define BIOHDF_H 1

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "hdf5.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/

/* fixed maximum data lengths, with lots of potential for trouble */
#define SEQID_LENGTH  100
#define REFID_LENGTH  1024
#define REFDEF_LENGTH 1024
#define SEQ_MAX_LENGTH 1024

#define REFERENCES_GROUP  "references"
#define SEQUENCES_GROUP   "sequences"
#define ALIGNMENTS_GROUP  "alignments"
#define SEQUENCE_DATASET  "sequence"
#define SEQID_DATASET     "seqid"
#define ALIGNMENT_DATASET "alignment"
#define REFMAP_DATASET    "refmap"
#define READMASK_DATASET  "readmask"
#define CLUSTER_DATASET   "cluster"
#define REFID_DATASET     "refid"
#define REFDEF_DATASET    "refdef"
#define REFIDX_DATASET    "refidx"
#define REFSEQ_DATASET    "refseq"


/* basic representation of an alignment hit */
typedef struct _hit_t {
  int read_id;
  int ref_id;
  int position;
  int length;
  int strand;
  int mismatch;
} hit_t;


/* used for indexing alignment hits, sorted by reference id and position */
typedef struct _refmap_t {
  int ref_id;
  int position;
  int length;
  int hit_id;
} refmap_t;


/* representation of a cluster of aligned reads - lookup into the alignment
   dataset would be required to recover the reads themselves */
typedef struct _cluster_t {
  int ref_id;
  int beg_pos;
  int end_pos;
  int num_read;
} cluster_t;


/* used for indexing into reference sequence 1D-char dataset */
typedef struct _refidx_t {
  long long offset;
  int length;
} refidx_t;


/******************************************************************************
 * Exported function declarations
 *****************************************************************************/

/* parse sequences from fasta format into sequence and seqid datasets */
int H5BIOparse_fastA(hid_t gid_seqs, int max_seq_length, int is_verbose);

/* parse sequences from fastq format into sequence and seqid datasets */
int H5BIOparse_fastQ(hid_t gid_seqs, int max_seq_length, int is_verbose);

/* create the hit memory datatype */
hid_t H5BIOhit_type();

/* create the refmap memory datatype */
hid_t H5BIOrefmap_type();

/* create the cluster memory datatype */
hid_t H5BIOcluster_type();

/* create the reference sequence index memory datatype */
hid_t H5BIOrefidx_type();

/* create the readmask dataset for this alignment */
int H5BIOreadmask_alignment(hid_t fid, hid_t gid_aln);

/* create the refmap index dataset for this alignment */
int H5BIOindex_alignment(hid_t fid, hid_t gid_aln);

void write_hit_buf(hid_t did_aln, int hit_num, const hit_t* hit_buf, int hit_buf_num, hid_t hit_tid);

int compare_refmap(const void *a, const void *b);

void refmap_item(refmap_t *item, hid_t did, hid_t sid, 
		 hid_t tid, hid_t mid, int pp);

int refmap_is_match(refmap_t *item, 
		    int ref_id, int start_pos, int end_pos, int seq_len);

int refmap_is_first(refmap_t *item, refmap_t *prior_item, 
		    int ref_id, int start_pos, int end_pos, int seq_len);

int refmap_is_prior(refmap_t *item, refmap_t *prior_item, 
		    int ref_id, int start_pos, int end_pos, int seq_len);

int refmap_is_after(refmap_t *item, refmap_t *prior_item, 
		    int ref_id, int start_pos, int end_pos, int seq_len);

/* binary search into an alignment set */
void refmap_binary_search(hid_t did, const char *ref_name, 
			  int ref_id, int start_pos, int end_pos,
			  int seq_len, 
			  void (*refmap_action)(const char *, const refmap_t *, int, int, void *),
			  void *info);

/* linear search over all hits in an alignment set */
void refmap_linear_search(hid_t did, const char *ref_name, 
			  int ref_id, int start_pos, int end_pos,
			  int start_id,	int seq_len, 
			  void (*refmap_action)(const char *, const refmap_t *, int, int, void *),
			  void *info);

void refmap_to_gff(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *info);

void refmap_to_depth(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *info);

void refmap_to_density(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *info);

/* write refidx entry into the dataset */
void H5BIOwrite_refidx(hid_t did, int num, const refidx_t* buf, int buf_num, hid_t tid);

#endif /* BIOHDF_H */
