/******************************************************************************
 *
 * BioHDF - HDF5-based data model and tools for storage and analysis of 
 *          Next Gen Sequencing reads and alignments. 
 *
 * Copyright (c) 2009 Geospiza, Inc.
 *
 * This program can be redistributed and/or modified under the terms of the 
 * Version 2 of the GNU General Public License as published by the Free 
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 *****************************************************************************/


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/

/* Script description */

const char *description = 
  "Export alignments to specified range of a reference sequence from dataset:\n"
  "    /alignments/<alignment>/alignment\n"
  "to text format (currently, gff).\n";


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *out_fmt, char *aln_name,
			  int *ref_id, int *ref_beg, int *ref_end, 
			  int *is_verbose, int *is_debug, int *is_force);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/


int main(int argc, char** argv) {
  
  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char out_fmt[40];
  char aln_name[40];
  int ref_id, ref_beg, ref_end;
  int is_verbose = 0;
  int is_debug   = 0;
  int is_force   = 0;

  hid_t fid = -1;
  hid_t gid = -1;
  hid_t did = -1;

  /* Option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(aln_name,     "alignment");
  strcpy(out_fmt,      "gff");
  ref_id  = 0;
  ref_beg = 0;
  ref_end = 300000000;

  /* Parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, out_fmt, aln_name,
			&ref_id, &ref_beg, &ref_end, 
			&is_verbose, &is_debug, &is_force) < 0);

  /* read from the HDF file */
  HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDWR, H5P_DEFAULT));

  /* open the requested alignment group */
  char hdf_path[PATH_MAX];
  sprintf(hdf_path, "/%s/%s", ALIGNMENTS_GROUP, aln_name);
  HDFERR(gid = H5Gopen(fid, hdf_path, H5P_DEFAULT));

  /* open the refmap dataset */
  GENERR(H5BIOindex_alignment(fid, gid));
  HDFERR(did = H5Dopen(gid, REFMAP_DATASET, H5P_DEFAULT));

  /* get fixed sequence length from 2nd dim of 'sequence' dataset extent */
  hid_t did_seq = -1;
  hid_t sid_seq = -1;
  sprintf(hdf_path, "/%s/%s", SEQUENCES_GROUP, SEQUENCE_DATASET);
  HDFERR(did_seq = H5Dopen(fid, hdf_path, H5P_DEFAULT));
  HDFERR(sid_seq = H5Dget_space(did_seq));
  int seq_len;
  H5UTLget_2D_extent(sid_seq, NULL, &seq_len);
  if (sid_seq >= 0) H5Sclose(sid_seq);
  if (did_seq >= 0) H5Dclose(did_seq);

  /* search for reads that align to specified region of a ref_id */
  if (strcmp(out_fmt, "gff") == 0) {
    printf("track name=hdf_output description=\"HDF Output\"\n");
    refmap_binary_search(did, aln_name, ref_id, ref_beg, ref_end, seq_len, 
			 refmap_to_gff, NULL);
  }
  else if (strcmp(out_fmt, "wig") == 0) {
    int i;
    int width = ref_end - ref_beg + 1;
    int *depth = (int *)calloc(width, sizeof(int));
    printf("track name=hdf_output description=\"HDF Output\"\n");
    printf("fixedStep chrom=%d start=%d step=1\n", ref_id, ref_beg);
    refmap_binary_search(did, aln_name, ref_id, ref_beg, ref_end, seq_len, 
			 refmap_to_depth, depth);
    for (i = 0; i < width; i++) {
      printf("%d\n", depth[i]);
    }
  }

  goto _no_error;
 _error:
  fprintf(stderr, "Fatal error, quitting...");
  return_code = 1;

 _no_error:

  /* tidy up HDF ids */
  if (did >= 0) H5Dclose(did);
  if (gid >= 0) H5Gclose(gid); 
  if (fid >= 0) H5Fclose(fid); 

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *out_fmt, char *aln_name,
			  int *ref_id, int *ref_beg, int *ref_end, 
			  int *is_verbose, int *is_debug, int *is_force) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:o:a:r:b:e:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument, is_verbose, 1},
    {"debug",    no_argument, is_debug, 1},
    {"force",    no_argument, is_force, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"out-fmt",  required_argument, 0, 'o'},
    {"aln",      required_argument, 0, 'a'},
    {"ref-id",   required_argument, 0, 'r'},
    {"ref-beg",  required_argument, 0, 'b'},
    {"ref-end",  required_argument, 0, 'e'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Help string */
  char *usage = "[--h5-file|-f <file.h5>] [--aln|-a <alignment>] \\\n"
    "        [--ref-id|-r 0] [--ref-beg|-b 0] [--ref-end|-e 300000000] \\\n"
    "        [--out-fmt|-o <gff>] \\\n"
    "        [--verbose] [--debug] [--force]"; 
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      if (optarg != NULL) strcpy(h5_file_name, optarg);
      break;
      
      /* output format */
    case 'o':
      if (optarg != NULL) strcpy(out_fmt, optarg);
      break;
      
      /* alignment name */
    case 'a':
      if (optarg != NULL) strcpy(aln_name, optarg);
      break;
      
      /* reference sequence internal id */
    case 'r':
      if (optarg != NULL) sscanf(optarg, "%d", ref_id);
      break;
      
      /* reference sequence range */
    case 'b':
      if (optarg != NULL) sscanf(optarg, "%d", ref_beg);
      break;
      
      /* reference sequence range */
    case 'e':
      if (optarg != NULL) sscanf(optarg, "%d", ref_end);
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      fprintf(stderr, "Usage:\n%s %s\n\nDescription:\n%s\n", argv[0], usage, description);
      exit(0);
    }
  }

  return 0;
}



