/******************************************************************************
 *
 * BioHDF - HDF5-based data model and tools for storage and analysis of 
 *          Next Gen Sequencing reads and alignments. 
 *
 * Copyright (c) 2009 Geospiza, Inc.
 *
 * This program can be redistributed and/or modified under the terms of the 
 * Version 2 of the GNU General Public License as published by the Free 
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 *****************************************************************************/


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/

/* Script description */
const char *description = 
  "Export file from a specified location in the HDF /files/ store to stdout.\n";


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, 
			  char *file_path,
			  int *is_verbose, int *is_debug, int *is_force);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char file_path[PATH_MAX];
  int is_verbose = 0;
  int is_debug   = 0;
  int is_force   = 0;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(file_path, "file.txt");

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, file_path, 
			&is_verbose, &is_debug, &is_force) < 0);

  /* open the HDF file */
  hid_t fid;
  if (file_exists(h5_file_name)) {
    HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDWR, H5P_DEFAULT));
  }
  else {
    fprintf(stderr, "File %s does not exist\n", h5_file_name);
    goto _error;
  }

  /* open the /files group that contains files as datasets */
  hid_t gid;
  if (H5UTLgroup_exists(fid, FILES_GROUP)) {
    gid = H5Gopen(fid, FILES_GROUP, H5P_DEFAULT);
  } else {
    fprintf(stderr, "Group /%s does not exist\n", FILES_GROUP);
    goto _error;
  }

  /* descend the groups hierarchy, mapping from path element to group */
  hid_t did;
  char *next_token;
  char *token = strtok(file_path, "/");
  for (;;) {
    if ((next_token = strtok(NULL, "/"))) {
      if (is_verbose) fprintf(stderr, "Opening group %s\n", token);
      if (H5UTLgroup_exists(gid, token)) {
	gid = H5Gopen(gid, token, H5P_DEFAULT);
      }
      else {
	fprintf(stderr, "Group %s does not exist\n", token);
	goto _error;
      }
      token = next_token;
    }
    else {
      /* last path element names the dataset to export */
      if (is_verbose) fprintf(stderr, "Open dataset %s\n", token);
      HDFERR(did = H5Dopen(gid, token, H5P_DEFAULT));
      break;
    }
  }

  /* buffered export of character dataset to stdout */
  char buf[HDF_RW_BUF_SIZE];
  int read_size, pos;
  for (pos = 0;
       (read_size = H5UTLread_1D_char_buf(did, pos, buf, HDF_RW_BUF_SIZE)) > 0;
       pos += read_size) {
    if (is_verbose) fprintf(stderr, "Reading %d bytes\n", read_size);
    write(1, buf, read_size);
  }


  goto _no_error;
 _error:
  fprintf(stderr, "Fatal error, quitting...\n");
  return_code = 1;

 _no_error:

  /* terminating process, so don't worry about tidying up */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *file_path, 
                          int *is_verbose, int *is_debug, int *is_force) {
  int c;
  int option_index = 0;

  /* short options */
  char *options = "f:p:h";

  /* long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"force",    no_argument,       is_force, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"path",     required_argument, 0, 'p'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* help string */
  char *usage = "[--h5-file|-f file.h5] --path|-p path/for/file.html \\\n"
    "        [--verbose] [--debug] [--force]"; 
  
  /* parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* if this option sets a flag, do nothing else now */
      break;
      
      /* h5 file */
    case 'f':
      if (optarg != NULL) strcpy(h5_file_name, optarg);
      break;
      
      /* path */
    case 'p':
      if (optarg != NULL) strcpy(file_path, optarg);
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      fprintf(stderr, "Usage:\n%s %s\n\nDescription:\n%s\n", argv[0], usage, description);
      exit(0);
    }
  }

  return 0;
}


