/******************************************************************************
 *
 * BioHDF - HDF5-based data model and tools for storage and analysis of 
 *          Next Gen Sequencing reads and alignments. 
 *
 * Copyright (c) 2009 Geospiza, Inc.
 *
 * This program can be redistributed and/or modified under the terms of the 
 * Version 2 of the GNU General Public License as published by the Free 
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 *****************************************************************************/


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/

/* Script description */
const char *description = 
  "Filter reads to leave those that have not hit against a list of \n"
  "alignment datasets and export those reads as fasta. \n"
  "Sequence identifier is the HDF index of the read, not the original id\n"
  "Alternatively, if --read_id|-d parameters are specified, only those \n"
  "reads are exported.\n";


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, 
			  char **filter_aln_names, int *num_filter, 
                          int **read_id, int *num_read_id, 
			  int *is_verbose, int *is_debug, int *is_force);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char *filter_aln_names[10]; /* TODO - should be unlimited */
  int num_filter;
  int *read_id = NULL;
  int num_read_id = 0;
  int is_verbose = 0;
  int is_debug   = 0;
  int is_force   = 0;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  num_filter = 0;

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, 
			filter_aln_names, &num_filter, 
			&read_id, &num_read_id, 
			&is_verbose, &is_debug, &is_force) < 0);

  /* open existing file using default properties */
  if (!file_exists(h5_file_name)) {
    fprintf(stderr, "File '%s' must exist\n", h5_file_name);
    goto _error;
  }
  hid_t fid;
  HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDWR, H5P_DEFAULT));

  /* reads must be available */
  if (!H5UTLgroup_exists(fid, SEQUENCES_GROUP)) {
    fprintf(stderr, "Group '%s' must exist before importing alignments; use 'h5_import_read'\n", SEQUENCES_GROUP);
    goto _error;
  }

  /* open the alignments group */
  hid_t gid_alns;
  HDFERR(gid_alns = H5Gopen(fid, ALIGNMENTS_GROUP, H5P_DEFAULT));

  /* get fixed sequence length from 2nd dim of 'sequence' dataset extent */
  char hdf_path[PATH_MAX];
  sprintf(hdf_path, "/%s/%s", SEQUENCES_GROUP, SEQUENCE_DATASET);
  hid_t did_seq, sid_seq;
  HDFERR(did_seq = H5Dopen(fid, hdf_path, H5P_DEFAULT));
  HDFERR(sid_seq = H5Dget_space(did_seq));
  int seq_num, seq_len;
  H5UTLget_2D_extent(sid_seq, &seq_num, &seq_len);

  /* form the OR-ed read-mask */
  char *full_readmask = (char *)calloc(seq_num, sizeof(char));
  /* optionally add a readmask from explicit read_id list passed */
  if (num_read_id > 0) {
    int i, j;
    for (j = 0; j < seq_num; j++) {
      full_readmask[j] = 1;
    }
    for (i = 0; i < num_read_id; i++) {
      if (0 <= read_id[i] && read_id[i] < seq_num) {
	full_readmask[read_id[i]] = 0;
      }
    }
  }
  char *one_readmask = (char *)calloc(seq_num, sizeof(char));
  int i, j;
  hid_t gid_aln, did_mask;
  for (i = 0; i < num_filter; i++) {
    HDFERR(gid_aln = H5Gopen(gid_alns, filter_aln_names[i], H5P_DEFAULT));
    GENERR(H5BIOreadmask_alignment(fid, gid_aln));
    HDFERR(did_mask = H5Dopen(gid_aln, READMASK_DATASET, H5P_DEFAULT));
    H5Dread(did_mask, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, one_readmask);
    for (j = 0; j < seq_num; j++) {
      full_readmask[j] |= one_readmask[j];
    }
    H5Dclose(did_mask);
    H5Gclose(gid_aln);
  }

  /* filter the reads using the composite read-mask */
  int buf_size = HDF_RW_BUF_SIZE;
  char *seq_buf = (char *)calloc(buf_size * seq_len, sizeof(char));

  /* iterate over all sequences */
  char *seq = (char *)calloc(seq_len + 1, sizeof(char));
  int buf_offset = - buf_size;
  for (i = 0; i < seq_num; i++) {
    if (i >= buf_offset + buf_size) {
      /* fill buffer */
      buf_offset += buf_size;
      int buf_fill = seq_num - buf_offset;
      if (buf_fill > buf_size) {
	buf_fill = buf_size;
      }
      hsize_t slab_start[2] = { buf_offset, 0 };
      hsize_t slab_count[2] = { buf_fill, seq_len };
      H5Sselect_hyperslab(sid_seq, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(2, slab_count, slab_count); 
      H5Dread(did_seq, H5T_NATIVE_CHAR, mid, sid_seq, H5P_DEFAULT, seq_buf);
      H5Sclose(mid);
    }
    memcpy(seq, seq_buf + (i - buf_offset) * seq_len, seq_len);
    *(seq + seq_len) = 0;

    if (full_readmask[i] == 0) {
      printf(">%d\n%s\n", i, seq);
    }
  }


  goto _no_error;
 _error:
  fprintf(stderr, "Fatal error, quitting...\n");
  return_code = 1;

 _no_error:

  /* tidy up HDF ids */

  /* terminating process, so don't worry about freeing the heap */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, 
			  char **filter_aln_names, int *num_filter, 
			  int **read_id, int *num_read_id, 
			  int *is_verbose, int *is_debug, int *is_force) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:a:o:x:d:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"force",    no_argument,       is_force, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"exclude",  required_argument, 0, 'x'},
    {"read_id",  required_argument, 0, 'd'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Help string */
  char *usage = "[--h5-file|-f <file.h5>] [--exclude|-x <mask_alignment>]+ \\\n"
    "        [--read_id|-d <read_id>]+ \\\n"
    "        [--verbose] [--debug] [--force]";
  
  int size_read_id = 0;

  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      if (optarg != NULL) strcpy(h5_file_name, optarg);
      break;
      
      /* alignment name for masking of reads */
    case 'x':
      if (optarg != NULL) {
	filter_aln_names[*num_filter] = (char *)calloc(HDF_OBJNAME_MAX, sizeof(char));
	strcpy(filter_aln_names[*num_filter], optarg);
	(*num_filter)++;
      }
      break;
      
      /* read_id list - used as a filter if it exists */
    case 'd':
      if (optarg != NULL) {
	int a_read_id;
	sscanf(optarg, "%d", &a_read_id);
	if (*read_id == NULL) {
	  size_read_id = 1;
	  *read_id = (int *)malloc(size_read_id * sizeof(int));
	}
	else if (*num_read_id == size_read_id) {
	  size_read_id *= 2;
	  *read_id = (int *)realloc(*read_id, size_read_id * sizeof(int));
	}
	(*read_id)[*num_read_id] = a_read_id;
	(*num_read_id)++;
      }
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      fprintf(stderr, "Usage:\n%s %s\n\nDescription:\n%s\n", argv[0], usage, description);
      exit(0);
    }
  }

  return 0;
}


