/******************************************************************************
 *
 * BioHDF - HDF5-based data model and tools for storage and analysis of 
 *          Next Gen Sequencing reads and alignments. 
 *
 * Copyright (c) 2009 Geospiza, Inc.
 *
 * This program can be redistributed and/or modified under the terms of the 
 * Version 2 of the GNU General Public License as published by the Free 
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 *****************************************************************************/


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/

/* Script description */
const char *description = 
  "Export specified region of reference sequence.\n";


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *ref_name, 
			  int *ref_id, int *ref_beg, int *ref_end, 
			  int *is_verbose, int *is_debug, int *is_force);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char ref_name[HDF_OBJNAME_MAX];
  int ref_id, ref_beg, ref_end;
  int is_verbose = 0;
  int is_debug   = 0;
  int is_force   = 0;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(ref_name, "refdb");
  ref_id  = 0;
  ref_beg = 0;
  ref_end = 300000000;

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, ref_name, 
			&ref_id, &ref_beg, &ref_end, 
			&is_verbose, &is_debug, &is_force) < 0);

  /* open existing file using default properties */
  if (!file_exists(h5_file_name)) {
    fprintf(stderr, "File '%s' must exist\n", h5_file_name);
    goto _error;
  }
  hid_t fid;
  HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDWR, H5P_DEFAULT));

  /* references must be available */
  if (!H5UTLgroup_exists(fid, REFERENCES_GROUP)) {
    fprintf(stderr, "Group '%s' must exist; use 'h5_import_reference'\n", REFERENCES_GROUP);
    goto _error;
  }

  /* open the reference group */
  hid_t gid_refs, gid_ref;
  HDFERR(gid_refs = H5Gopen(fid, REFERENCES_GROUP, H5P_DEFAULT));
  HDFERR(gid_ref = H5Gopen(gid_refs, ref_name, H5P_DEFAULT));

  /* read sequence index dataset to memory */
  hid_t did_refidx, tid_refidx, sid_refidx;
  HDFERR(did_refidx = H5Dopen(gid_ref, REFIDX_DATASET, H5P_DEFAULT));
  HDFERR(sid_refidx = H5Dget_space(did_refidx));
  int refidx_num = H5UTLget_1D_extent(sid_refidx);
  refidx_t *refidx = (refidx_t *)calloc(refidx_num, sizeof(refidx_t));
  HDFERR(tid_refidx = H5BIOrefidx_type());
  H5Dread(did_refidx, tid_refidx, H5S_ALL, H5S_ALL, H5P_DEFAULT, refidx);
  
  long long refseq_offset = (*(refidx + ref_id)).offset;
  int refseq_length = (*(refidx + ref_id)).length;

  long long req_offset = refseq_offset + (long long)ref_beg;
  int req_length = ref_end - ref_beg;

  if (req_length <= 0 || 
      req_offset + req_length > refseq_offset + refseq_length) {
    fprintf(stderr, "Bad parameters");
    exit(1);
  }

  /* select 1D char from dataset refseq in range req_offset to req_length */
  hid_t did_refseq, sid_refseq;
  HDFERR(did_refseq = H5Dopen(gid_ref, REFSEQ_DATASET, H5P_DEFAULT));
  HDFERR(sid_refseq = H5Dget_space(did_refseq));
  char *refseq = (char *)calloc(req_length + 1, sizeof(char));
  HDFRET(sid_refseq = H5Dget_space(did_refseq));
  hsize_t slab_start[1] = { req_offset };
  hsize_t slab_count[1] = { req_length };
  H5Sselect_hyperslab(sid_refseq, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
  hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
  H5Dread(did_refseq, H5T_NATIVE_CHAR, mid, sid_refseq, H5P_DEFAULT, refseq);
  H5Sclose(mid);

  *(refseq + req_length) = '\0';

  printf("%s\n", refseq);


  goto _no_error;
 _error:
  fprintf(stderr, "Fatal error, quitting...\n");
  return_code = 1;

 _no_error:

  /* tidy up HDF ids */

  /* terminating process, so don't worry about freeing the heap */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *ref_name, 
			  int *ref_id, int *ref_beg, int *ref_end, 
			  int *is_verbose, int *is_debug, int *is_force) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:d:r:b:e:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"force",    no_argument,       is_force, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"ref",      required_argument, 0, 'd'},
    {"ref-id",   required_argument, 0, 'r'},
    {"ref-beg",  required_argument, 0, 'b'},
    {"ref-end",  required_argument, 0, 'e'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Help string */
  char *usage = "[--h5-file|-f <file.h5>] [--ref|-d <ref_name>] \\\n"
    "        [--ref-id|-r 0] [--ref-beg|-b 0] [--ref-end|-e 300000000] \\\n"
    "        [--verbose] [--debug] [--force]"; 

  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      if (optarg != NULL) strcpy(h5_file_name, optarg);
      break;
      
      /* reference database */
    case 'd':
      if (optarg != NULL) strcpy(ref_name, optarg);
      break;
      
      /* reference sequence internal id */
    case 'r':
      if (optarg != NULL) sscanf(optarg, "%d", ref_id);
      break;
      
      /* reference sequence range */
    case 'b':
      if (optarg != NULL) sscanf(optarg, "%d", ref_beg);
      break;
      
      /* reference sequence range */
    case 'e':
      if (optarg != NULL) sscanf(optarg, "%d", ref_end);
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      fprintf(stderr, "Usage:\n%s %s\n\nDescription:\n%s\n", argv[0], usage, description);
      exit(0);
    }
  }

  return 0;
}


