/******************************************************************************
 *
 * BioHDF - HDF5-based data model and tools for storage and analysis of 
 *          Next Gen Sequencing reads and alignments. 
 *
 * Copyright (c) 2009 Geospiza, Inc.
 *
 * This program can be redistributed and/or modified under the terms of the 
 * Version 2 of the GNU General Public License as published by the Free 
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 *****************************************************************************/


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/

/* Script description */
const char *description = 
  "Import alignment of reads against a reference in mapreads (.csfasta) or\n"
  "SAM format to dataset:\n"
  "    /alignments/<alignment>/alignment\n";


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *in_fmt, char *aln_name, 
			  int *is_verbose, int *is_debug, int *is_force);

static int _parse_mapreads(hid_t did_aln, hid_t tid_hit, int *read_len);

static int _parse_bowtie(hid_t did_aln, hid_t tid_hit, int *read_len);

static int _parse_sam(hid_t did_aln, hid_t tid_hit, int *read_len);

static int _parse_bowtie_extended(hid_t did_aln, hid_t tid_hit, int *read_len);

static int _optNM_to_mismatch(char *opt);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char in_fmt[40];
  char aln_name[HDF_OBJNAME_MAX];
  int is_verbose = 0;
  int is_debug   = 0;
  int is_force   = 0;

  hid_t fid        = -1;
  hid_t gid_alns   = -1;
  hid_t gid_aln    = -1;
  hid_t did_aln    = -1;
  hid_t tid_hit    = -1;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(in_fmt,       "mapreads");
  strcpy(aln_name,     "alignment");

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, in_fmt, aln_name, 
			&is_verbose, &is_debug, &is_force) < 0);

  /* append to the existing file using default properties */
  if (!file_exists(h5_file_name)) {
    fprintf(stderr, "File '%s' must exist before importing alignments; use 'h5_import_read'\n", h5_file_name);
    goto _error;
  }
  HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDWR, H5P_DEFAULT));

  /* reads must be imported before alignment annotations added */
  if (!H5UTLgroup_exists(fid, SEQUENCES_GROUP)) {
    fprintf(stderr, "Group '%s' must exist before importing alignments; use 'h5_import_read'\n", SEQUENCES_GROUP);
    goto _error;
  }

  /* create the alignments group, if not already present */
  HDFERR(gid_alns = H5UTLopen_or_create_group(fid, ALIGNMENTS_GROUP));

  /* open named alignment group */
  if (H5UTLgroup_exists(gid_alns, aln_name)) {
    fprintf(stderr, "Group '/%s/%s' already exists, not overwriting\n", ALIGNMENTS_GROUP, aln_name);
    goto _error;
  }
  HDFERR(gid_aln = H5Gcreate(gid_alns, aln_name, H5P_DEFAULT, H5P_DEFAULT, 0));

  /* create the hit memory datatype */
  HDFERR(tid_hit = H5BIOhit_type());

  /* create the "alignment" dataset (list of individual read/ref hits) */
  HDFERR(did_aln = H5UTLcreate_1D_dataset(gid_aln, "alignment", tid_hit));

  /* open the "sequence" dataset for the total 'reads' statistic */
  hid_t gid_seqs, did_seq, sid_seq;
  HDFERR(gid_seqs = H5Gopen(fid, SEQUENCES_GROUP, H5P_DEFAULT));
  HDFERR(did_seq = H5Dopen(gid_seqs, SEQUENCE_DATASET, H5P_DEFAULT));
  HDFERR(sid_seq = H5Dget_space(did_seq));
  int seq_num, seq_len;
  H5UTLget_2D_extent(sid_seq, &seq_num, &seq_len);
  
  /* store the 'num_read' statistic */
  H5UTLwrite_attr_int(did_aln, "num_read", seq_num);

  /* Get the read lengths */
  int *read_len = (int *)calloc(seq_num, sizeof(int));
  H5UTLget_2D_char_len(did_seq, read_len, seq_num);

  /* For now, split the mapreads/csfasta parser from the SAM one. Lots of 
     common code can be abstracted out later. */
  if (strcmp(in_fmt, "mapreads") == 0) {
    GENERR(_parse_mapreads(did_aln, tid_hit, read_len));
  }
  else if (strcmp(in_fmt, "bowtie") == 0) {
    GENERR(_parse_bowtie(did_aln, tid_hit, read_len));
  }
  else if (strcmp(in_fmt, "sam") == 0) {
    GENERR(_parse_sam(did_aln, tid_hit, read_len));
  }
  else if (strcmp(in_fmt, "bowtie_extended") == 0) {
    GENERR(_parse_bowtie_extended(did_aln, tid_hit, read_len));
  }

  goto _no_error;
 _error:
  fprintf(stderr, "Fatal error, quitting...\n");
  return_code = 1;

 _no_error:

  /* tidy up HDF ids */
  if (tid_hit    >= 0) H5Tclose(tid_hit);
  if (did_aln    >= 0) H5Dclose(did_aln);
  if (gid_aln    >= 0) H5Gclose(gid_aln); 
  if (gid_alns   >= 0) H5Gclose(gid_alns); 
  if (fid        >= 0) H5Fclose(fid); 

  /* terminating process, so don't worry about freeing the heap */
  /* free(hit_buf); */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *in_fmt, char *aln_name, 
                          int *is_verbose, int *is_debug, int *is_force) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:i:a:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"force",    no_argument,       is_force, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"in-fmt",   required_argument, 0, 'i'},
    {"aln",      required_argument, 0, 'a'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Help string */
  char *usage = "[--h5-file|-f <file.h5>] [--aln|-a <alignment>] \\\n"
    "        [--in-fmt|-i mapreads|bowtie|sam|gff] \\\n"
    "        [--verbose] [--debug] [--force]"; 
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      if (optarg != NULL) strcpy(h5_file_name, optarg);
      break;
      
      /* input format */
    case 'i':
      if (optarg != NULL) strcpy(in_fmt, optarg);
      break;
      
      /* alignment name */
    case 'a':
      if (optarg != NULL) strcpy(aln_name, optarg);
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      fprintf(stderr, "Usage:\n%s %s\n\nDescription:\n%s\n", argv[0], usage, description);
      exit(0);
    }
  }

  return 0;
}


/* parse alignments in mapreads/csfasta format */
static int _parse_mapreads(hid_t did_aln, hid_t tid_hit, int *read_len) {

  char line[1024];
  char seqid_hits[400];
  int seq_num = 0;
  int hit_num = 0;
  int hit_buf_size = HDF_RW_BUF_SIZE;
  hit_t *hit_buf = (hit_t *)calloc(hit_buf_size, sizeof(hit_t));
  hit_t hit;

  while (fgetline(stdin, line, sizeof(line)) > 0) {

    /* skip comments */
    if (line[0] == '#') {
      continue;
    }
    if (strncmp(line, "map start", 9) == 0) {
      continue;
    }

    /* parse line into read ID and ref_id hits */
    if (sscanf(line, ">%s", seqid_hits) == 1) {

      char *token = strtok(seqid_hits, ",");
      while ((token = strtok(NULL, ","))) {
	/* ignore the read ID (assumes order is the same as the original csfasta file - big assumption!) */

	int ref_id, position, mismatch;
	if (sscanf(token, "%d_%d.%d", &ref_id, &position, &mismatch) == 3) {
	} else if (sscanf(token, "%d.%d", &position, &mismatch) == 2) {
	  ref_id = 0;
	} else {
	  fprintf(stderr, "File format problem:\n%s\n", line);
	  exit(1);
	}
	
	hit.read_id  = seq_num;
	hit.ref_id   = ref_id;
	hit.position = position;
	hit.length   = read_len[hit.read_id];
	hit.strand   = 1;
	if (hit.position < 0) {
	  hit.strand = -1;
	  hit.position = abs(hit.position);
	}
	hit.mismatch = mismatch;
	
	/* append the hit to dataset, via a buffer */
	int buf_loc = hit_num % hit_buf_size;
	memcpy(hit_buf + buf_loc, &hit, sizeof(hit_t));
	++hit_num;
	if (buf_loc == hit_buf_size - 1) {
	  write_hit_buf(did_aln, hit_num, hit_buf, hit_buf_size, tid_hit);
	}
      }
    }
    ++seq_num;
  }
  
  /* append last buffered hits to dataset */
  int buf_loc = hit_num % hit_buf_size;
  if (buf_loc > 0) {
    write_hit_buf(did_aln, hit_num, hit_buf, buf_loc, tid_hit);
  }

  return 0;
}


/* parse alignments in bowtie/concise format */
static int _parse_bowtie(hid_t did_aln, hid_t tid_hit, int *read_len) {

  char line[1024];
  int hit_num = 0;
  int hit_buf_size = HDF_RW_BUF_SIZE;
  hit_t *hit_buf = (hit_t *)calloc(hit_buf_size, sizeof(hit_t));
  hit_t hit;

  while (fgetline(stdin, line, sizeof(line)) > 0) {

    /* parse line into read ID and ref_id hits */
    int read_id, ref_id, position, mismatch;
    char strand;
    if (sscanf(line, "%d%c:<%d,%d,%d>", 
	       &read_id, &strand, &ref_id, &position, &mismatch) == 5) {

      hit.read_id  = read_id; /* 0-based index */
      hit.ref_id   = ref_id; /* 0-based index */
      hit.position = position; /* 0-based position */
      hit.length   = read_len[hit.read_id];
      hit.strand   = 1;
      if (strand == '-') {
	hit.strand = -1;
      }
      hit.mismatch = mismatch;
	
      /* append the hit to dataset, via a buffer */
      int buf_loc = hit_num % hit_buf_size;
      memcpy(hit_buf + buf_loc, &hit, sizeof(hit_t));
      ++hit_num;
      if (buf_loc == hit_buf_size - 1) {
	write_hit_buf(did_aln, hit_num, hit_buf, hit_buf_size, tid_hit);
      }
    }
  }
  
  /* append last buffered hits to dataset */
  int buf_loc = hit_num % hit_buf_size;
  if (buf_loc > 0) {
    write_hit_buf(did_aln, hit_num, hit_buf, buf_loc, tid_hit);
  }

  return 0;
}


/* parse alignments in SAM format */
static int _parse_sam(hid_t did_aln, hid_t tid_hit, int *read_len) {

  char line[1024];
  int hit_num = 0;
  int hit_buf_size = HDF_RW_BUF_SIZE;
  hit_t *hit_buf = (hit_t *)calloc(hit_buf_size, sizeof(hit_t));
  hit_t hit;

  while (fgetline(stdin, line, sizeof(line)) > 0) {

    /* skip headers */
    if (line[0] == '@') {
      continue;
    }

    /* parse line into read ID and ref_id hits */
    int qname, flag, rname, pos, mapq, mpos, isize;
    char cigar[200];
    char mrnm[80];
    char seq[1024];
    char qual[1024];
    char opt1[1024];
    char opt2[1024];
    
    if (sscanf(line, "%d\t%d\t%d\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s\t%s\t%s", &qname, &flag, &rname, &pos, &mapq, cigar, mrnm, &mpos, &isize, seq, qual, opt1, opt2) == 13) {

      if (flag & 0x4) continue;

      hit.read_id  = qname;
      hit.ref_id   = rname;
      hit.position = pos - 1; /* to 0-based */
      hit.length   = read_len[hit.read_id];
      hit.strand   = 1;
      if (flag & 0x10) hit.strand = -1;
      hit.mismatch = _optNM_to_mismatch(opt2); /* TODO may be positioned differently - this needs to be more flexible */
	
      /* append the hit to dataset, via a buffer */
      int buf_loc = hit_num % hit_buf_size;
      memcpy(hit_buf + buf_loc, &hit, sizeof(hit_t));
      ++hit_num;
      if (buf_loc == hit_buf_size - 1) {
	write_hit_buf(did_aln, hit_num, hit_buf, hit_buf_size, tid_hit);
      }
    }
  }
  
  /* append last buffered hits to dataset */
  int buf_loc = hit_num % hit_buf_size;
  if (buf_loc > 0) {
    write_hit_buf(did_aln, hit_num, hit_buf, buf_loc, tid_hit);
  }

  return 0;
}


/* parse alignments in an extended bowtie/concise format */
static int _parse_bowtie_extended(hid_t did_aln, hid_t tid_hit, int *read_len) {

  char line[1024];
  int hit_num = 0;
  int hit_buf_size = HDF_RW_BUF_SIZE;
  hit_t *hit_buf = (hit_t *)calloc(hit_buf_size, sizeof(hit_t));
  hit_t hit;

  while (fgetline(stdin, line, sizeof(line)) > 0) {

    /* parse line into read ID and ref_id hits */
    int read_id, ref_id, position, mismatch, length;
    char strand;
    if (sscanf(line, "%d%c:<%d,%d,%d,%d>", 
	       &read_id, &strand, &ref_id, &position, &mismatch, &length) == 6) {

      hit.read_id  = read_id; /* 0-based index */
      hit.ref_id   = ref_id; /* 0-based index */
      hit.position = position; /* 0-based position */
      hit.length   = length;
      hit.strand   = 1;
      if (strand == '-') {
	hit.strand = -1;
      }
      hit.mismatch = mismatch;
	
      /* append the hit to dataset, via a buffer */
      int buf_loc = hit_num % hit_buf_size;
      memcpy(hit_buf + buf_loc, &hit, sizeof(hit_t));
      ++hit_num;
      if (buf_loc == hit_buf_size - 1) {
	write_hit_buf(did_aln, hit_num, hit_buf, hit_buf_size, tid_hit);
      }
    }
  }
  
  /* append last buffered hits to dataset */
  int buf_loc = hit_num % hit_buf_size;
  if (buf_loc > 0) {
    write_hit_buf(did_aln, hit_num, hit_buf, buf_loc, tid_hit);
  }

  return 0;
}


/* number of mismatches */
static int _optNM_to_mismatch(char *opt) {
  int mismatch = 0;
  if (sscanf(opt, "NM:i:%d", &mismatch) == 1) return mismatch;
  return 0;
}
