/******************************************************************************
 *
 * BioHDF - HDF5-based data model and tools for storage and analysis of 
 *          Next Gen Sequencing reads and alignments. 
 *
 * Copyright (c) 2009 Geospiza, Inc.
 *
 * This program can be redistributed and/or modified under the terms of the 
 * Version 2 of the GNU General Public License as published by the Free 
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 *****************************************************************************/


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <string.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/

/* Script description */
const char *description = 
  "Import a sequence file from csfasta, fasta, or fastq to HDF5 datasets:\n"
  "    /sequences/seqid - 2D char of sequence identifiers\n"
  "    /sequences/sequence - 2D char of sequences\n"
  "If no parameter --seq-len -l is provided, the maximum sequence length\n"
  "is taken to be the length of the first read.\n";
/* TODO: qualities from fastq */
/*     /sequences/quality - 2D char of fastq-style ASCII-encoded qualities */

/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *in_fmt, int *max_seq_length, 
			  int *is_verbose, int *is_debug, int *is_force);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {
  
  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char in_fmt[40];
  int max_seq_length;
  int is_verbose = 0;
  int is_debug   = 0;
  int is_force   = 0;

  hid_t fid;
  hid_t gid_seqs;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(in_fmt, "fasta");
  max_seq_length = -1; /* implies get from first sequence */

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, in_fmt, &max_seq_length,
			&is_verbose, &is_debug, &is_force) < 0);

  /* create a new file using default properties */
  fid = H5UTLopen_or_create_file(h5_file_name);

  /* create the sequences group */
  if (H5UTLgroup_exists(fid, SEQUENCES_GROUP)) {
    fprintf(stderr, "Group '/%s' already exists, not overwriting\n", SEQUENCES_GROUP);
    goto _error;
  }
  HDFERR(gid_seqs = H5Gcreate(fid, SEQUENCES_GROUP, H5P_DEFAULT, H5P_DEFAULT, 0));
  H5Gset_comment(gid_seqs, ".", "Sequence collection");

  if (strcmp(in_fmt, "fasta") == 0 ||
      strcmp(in_fmt, "csfasta") == 0) {
    GENERR(H5BIOparse_fastA(gid_seqs, max_seq_length, is_verbose));
  } else if (strcmp(in_fmt, "fastq") == 0) {
    GENERR(H5BIOparse_fastQ(gid_seqs, max_seq_length, is_verbose));
  }
  

  goto _no_error;
 _error:
  fprintf(stderr, "Fatal error, quitting...\n");
  return_code = 1;

 _no_error:
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *in_fmt, int *max_seq_length,
                          int *is_verbose, int *is_debug, int *is_force) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:i:l:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"force",    no_argument,       is_force, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"in-fmt",   required_argument, 0, 'i'},
    {"seq-len",  required_argument, 0, 'l'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Help string */
  char *usage = "[--h5-file|-f <file.h5>] [--in-fmt|-i csfasta|fasta|fastq] \\\n"
    "        [--seq-len|-l <max_seq_length>] \\\n"
    "        [--verbose] [--debug] [--force]"; 
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      if (optarg != NULL) strcpy(h5_file_name, optarg);
      break;
      
      /* input format */
    case 'i':
      if (optarg != NULL) strcpy(in_fmt, optarg);
      break;
      
      /* max_seq_length (optional) */
    case 'l':
      if (optarg != NULL) sscanf(optarg, "%d", max_seq_length);
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      fprintf(stderr, "Usage:\n%s %s\n\nDescription:\n%s\n", argv[0], usage, description);
      exit(0);
    }
  }

  return 0;
}


