/******************************************************************************
 *
 * BioHDF - HDF5-based data model and tools for storage and analysis of 
 *          Next Gen Sequencing reads and alignments. 
 *
 * Copyright (c) 2009 Geospiza, Inc.
 *
 * This program can be redistributed and/or modified under the terms of the 
 * Version 2 of the GNU General Public License as published by the Free 
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 *****************************************************************************/


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/

/* Script description */
const char *description = 
  "Generate summary statistics for an alignment dataset:\n"
  "    /alignments/<alignment>/alignment\n";


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *aln_name, 
			  int *is_verbose, int *is_debug, int *is_force);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char aln_name[HDF_OBJNAME_MAX];
  int is_verbose = 0;
  int is_debug   = 0;
  int is_force   = 0;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(aln_name,     "alignment");

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, aln_name, 
			&is_verbose, &is_debug, &is_force) < 0);

  /* append to the existing file using default properties */
  if (!file_exists(h5_file_name)) {
    fprintf(stderr, "File '%s' must exist; use 'h5_import_read' and 'h5_import_alignment'\n", h5_file_name);
    goto _error;
  }
  hid_t fid;
  HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDWR, H5P_DEFAULT));

  /* reads must be imported before alignment annotations added */
  if (!H5UTLgroup_exists(fid, SEQUENCES_GROUP)) {
    fprintf(stderr, "Group '%s' must exist; use 'h5_import_read'\n", SEQUENCES_GROUP);
    goto _error;
  }

  /* alignment group must exist */
  if (!H5UTLgroup_exists(fid, ALIGNMENTS_GROUP)) {
    fprintf(stderr, "Group '/%s/%s' must exist; use 'h5_import_alignment'\n", ALIGNMENTS_GROUP, aln_name);
    goto _error;
  }
  hid_t gid_alns;
  HDFERR(gid_alns = H5Gopen(fid, ALIGNMENTS_GROUP, H5P_DEFAULT));

  /* open named alignment group */
  if (!H5UTLgroup_exists(gid_alns, aln_name)) {
    fprintf(stderr, "Group '/%s/%s' must exist; use 'h5_import_alignment'\n", ALIGNMENTS_GROUP, aln_name);
    goto _error;
  }
  hid_t gid_aln;
  HDFERR(gid_aln = H5Gopen(gid_alns, aln_name, H5P_DEFAULT));

  /* open the "alignment" dataset (list of individual read/ref hits) */
  hid_t did_aln;
  HDFERR(did_aln = H5Dopen(gid_aln, ALIGNMENT_DATASET, H5P_DEFAULT));

  /* create the hit memory datatype */
  hid_t tid_hit;
  HDFERR(tid_hit = H5BIOhit_type());

  /* iterate through the alignment dataset */
  hid_t sid_aln;
  HDFERR(sid_aln = H5Dget_space(did_aln));
  int hit_num = H5UTLget_1D_extent(sid_aln);

  /* use fixed-size buffer */
  int buf_size = HDF_RW_BUF_SIZE;
  hit_t *hit_buf = (hit_t *)malloc(sizeof(hit_t) * buf_size);

  /* count of reads mapped */
  int num_read_mapped = 0;
  int last_read_id = -1;
  /* count mapped with 0,1,2,... mismatch */
  int num_read_mapped_0 = 0;
  int num_read_mapped_1 = 0;
  int num_read_mapped_2 = 0;
  int num_read_mapped_3 = 0;
  int num_read_mapped_4 = 0;
  int num_read_mapped_5 = 0;
  int num_read_mapped_6m = 0;
  int least_mismatch = -1;

  /* iterate over all hits */
  int i;
  hit_t *item;
  int buf_offset = - buf_size;
  for (i = 0; i < hit_num; i++) {
    if (i >= buf_offset + buf_size) {
      /* fill buffer */
      buf_offset += buf_size;
      int buf_fill = hit_num - buf_offset;
      if (buf_fill > buf_size) {
	buf_fill = buf_size;
      }
      hsize_t slab_start[1] = { buf_offset };
      hsize_t slab_count[1] = { buf_fill };
      H5Sselect_hyperslab(sid_aln, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
      H5Dread(did_aln, tid_hit, mid, sid_aln, H5P_DEFAULT, hit_buf);
      H5Sclose(mid);
    }
    item = hit_buf + i - buf_offset;

    if (last_read_id != item->read_id) {
      switch (least_mismatch) {
      case -1:                       break;
      case 0:  ++num_read_mapped_0;  break;
      case 1:  ++num_read_mapped_1;  break;
      case 2:  ++num_read_mapped_2;  break;
      case 3:  ++num_read_mapped_3;  break;
      case 4:  ++num_read_mapped_4;  break;
      case 5:  ++num_read_mapped_5;  break;
      default: ++num_read_mapped_6m; break;
      }
      least_mismatch = item->mismatch;

      last_read_id = item->read_id;
      ++num_read_mapped;
    } else {
      if (least_mismatch > item->mismatch) least_mismatch = item->mismatch;
    }
  }
  switch (least_mismatch) {
  case -1:                       break;
  case 0:  ++num_read_mapped_0;  break;
  case 1:  ++num_read_mapped_1;  break;
  case 2:  ++num_read_mapped_2;  break;
  case 3:  ++num_read_mapped_3;  break;
  case 4:  ++num_read_mapped_4;  break;
  case 5:  ++num_read_mapped_5;  break;
  default: ++num_read_mapped_6m; break;
  }
  
  /* store the 'num_read_mapped' statistic */
  H5UTLwrite_attr_int(did_aln, "num_read_mapped", num_read_mapped);

  /* store the 'num_read_mapped_n' mismatch statistics */
  H5UTLwrite_attr_int(did_aln, "num_read_mapped_0", num_read_mapped_0);
  H5UTLwrite_attr_int(did_aln, "num_read_mapped_1", num_read_mapped_1);
  H5UTLwrite_attr_int(did_aln, "num_read_mapped_2", num_read_mapped_2);
  H5UTLwrite_attr_int(did_aln, "num_read_mapped_3", num_read_mapped_3);
  H5UTLwrite_attr_int(did_aln, "num_read_mapped_4", num_read_mapped_4);
  H5UTLwrite_attr_int(did_aln, "num_read_mapped_5", num_read_mapped_5);
  H5UTLwrite_attr_int(did_aln, "num_read_mapped_6", num_read_mapped_6m);

  goto _no_error;
 _error:
  fprintf(stderr, "Fatal error, quitting...\n");
  return_code = 1;

 _no_error:

  /* terminating process, so don't worry about tidying up */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *aln_name, 
                          int *is_verbose, int *is_debug, int *is_force) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:a:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"force",    no_argument,       is_force, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"aln",      required_argument, 0, 'a'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Help string */
  char *usage = "[--h5-file|-f <file.h5>] [--aln|-a <alignment>] \\\n"
    "        [--verbose] [--debug] [--force]"; 
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      if (optarg != NULL) strcpy(h5_file_name, optarg);
      break;
      
      /* alignment name */
    case 'a':
      if (optarg != NULL) strcpy(aln_name, optarg);
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      fprintf(stderr, "Usage:\n%s %s\n\nDescription:\n%s\n", argv[0], usage, description);
      exit(0);
    }
  }

  return 0;
}


