/******************************************************************************
 *
 * BioHDF - HDF5-based data model and tools for storage and analysis of 
 *          Next Gen Sequencing reads and alignments. 
 *
 * Copyright (c) 2009 Geospiza, Inc.
 *
 * This program can be redistributed and/or modified under the terms of the 
 * Version 2 of the GNU General Public License as published by the Free 
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 *****************************************************************************/


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/

/* Script description */
const char *description = 
  "Intersect specified read-masks, count items and output.\n";  


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, 
			  char **filter_aln_names, int *num_filter, 
			  int *is_verbose, int *is_debug, int *is_force);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char *filter_aln_names[10];
  int num_filter;
  int is_verbose = 0;
  int is_debug   = 0;
  int is_force   = 0;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  num_filter = 0;

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, 
			filter_aln_names, &num_filter, 
			&is_verbose, &is_debug, &is_force) < 0);

  /* check for file */
  if (!file_exists(h5_file_name)) {
    fprintf(stderr, "File '%s' must exist\n", h5_file_name);
    goto _error;
  }
  hid_t fid;
  HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDWR, H5P_DEFAULT));

  /* open the alignments group */
  hid_t gid_alns;
  HDFERR(gid_alns = H5Gopen(fid, ALIGNMENTS_GROUP, H5P_DEFAULT));

  /* get 'sequence' dataset extent */
  char hdf_path[PATH_MAX];
  sprintf(hdf_path, "/%s/%s", SEQUENCES_GROUP, SEQUENCE_DATASET);
  hid_t did_seq, sid_seq;
  HDFERR(did_seq = H5Dopen(fid, hdf_path, H5P_DEFAULT));
  HDFERR(sid_seq = H5Dget_space(did_seq));
  int num_seq, seq_len;
  H5UTLget_2D_extent(sid_seq, &num_seq, &seq_len);
  H5Sclose(sid_seq);
  H5Dclose(did_seq);

  /* form the OR-ed read-mask */
  char *full_readmask = (char *)calloc(num_seq, sizeof(char));
  char *one_readmask = (char *)calloc(num_seq, sizeof(char));
  int i, j;
  hid_t gid_aln, did_mask;
  for (i = 0; i < num_filter; i++) {
    HDFERR(gid_aln = H5Gopen(gid_alns, filter_aln_names[i], H5P_DEFAULT));
    GENERR(H5BIOreadmask_alignment(fid, gid_aln));
    HDFERR(did_mask = H5Dopen(gid_aln, READMASK_DATASET, H5P_DEFAULT));
    H5Dread(did_mask, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, one_readmask);
    for (j = 0; j < num_seq; j++) {
      full_readmask[j] |= one_readmask[j];
    }
    H5Dclose(did_mask);
    H5Gclose(gid_aln);
  }

  /* output the remaining unmasked reads */
  int num_seq_filter = 0;
  for (j = 0; j < num_seq; j++) {
    if (full_readmask[j] == 0) ++num_seq_filter;
  }
  printf("num_read\t%d\n", num_seq_filter);

  goto _no_error;
 _error:
  fprintf(stderr, "Fatal error, quitting...\n");
  return_code = 1;

 _no_error:

  /* tidy up HDF ids */

  /* terminating process, so don't worry about freeing the heap */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, 
			  char **filter_aln_names, int *num_filter, 
			  int *is_verbose, int *is_debug, int *is_force) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:a:o:x:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"force",    no_argument,       is_force, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"exclude",  required_argument, 0, 'x'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Help string */
  char *usage = "[--h5-file|-f <file.h5>] [--exclude|-x <alignment>]+ \\\n"
    "        [--verbose] [--debug] [--force]"; 
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      if (optarg != NULL) strcpy(h5_file_name, optarg);
      break;
      
      /* alignment name for masking of reads */
    case 'x':
      if (optarg != NULL) {
	filter_aln_names[*num_filter] = (char *)calloc(HDF_OBJNAME_MAX, sizeof(char));
	strcpy(filter_aln_names[*num_filter], optarg);
	(*num_filter)++;
      }
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      fprintf(stderr, "Usage:\n%s %s\n\nDescription:\n%s\n", argv[0], usage, description);
      exit(0);
    }
  }

  return 0;
}


