/******************************************************************************
 *
 * BioHDF - HDF5-based data model and tools for storage and analysis of 
 *          Next Gen Sequencing reads and alignments. 
 *
 * Copyright (c) 2009 Geospiza, Inc.
 *
 * This program can be redistributed and/or modified under the terms of the 
 * Version 2 of the GNU General Public License as published by the Free 
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 *****************************************************************************/


/******************************************************************************
 *
 * HDF5 utilities used in the BioHDF project. 
 *
 *****************************************************************************/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

#include "hdf5.h"
#include "hdfutil.h"


int fgetline(FILE *pf, char s[], int lim) {
  int c = 0;
  int i = 0;
  while (--lim > 0 && (c = fgetc(pf)) != EOF && c != '\n') {
    s[i++] = c;
  }
  if (c == '\n') {
    s[i++] = c;
  }
  s[i] = '\0';
  return i;
}


int file_exists (const char * file_name) {
  struct stat buf;
  int i = stat ( file_name, &buf );
  if (i == 0) {
    return 1;
  }
  return 0;
}


/* structure for _object_exists */
struct _has_object_data {
  int has;
  char name[HDF_OBJNAME_MAX];
  int type;
};

/* callback function for _object_exists */
herr_t _has_object(hid_t loc_id, const char *name, void *opdata)
{
  struct _has_object_data *has_object_data = (struct _has_object_data *)opdata;
  H5G_stat_t statbuf;
  H5Gget_objinfo(loc_id, name, 0, &statbuf);
  if (statbuf.type == has_object_data->type &&
      strcmp(name, has_object_data->name) == 0) {
    has_object_data->has = 1;
  }
  return 0;
}

int _object_exists(hid_t loc_id, const char *name, int type) {
  struct _has_object_data has_object_data;
  has_object_data.has = 0;
  strcpy(has_object_data.name, name);
  has_object_data.type = type;
  H5Giterate(loc_id, ".", NULL, _has_object, &has_object_data);
  return has_object_data.has;
}


int H5UTLgroup_exists(hid_t loc_id, const char *name) {
  return _object_exists(loc_id, name, H5G_GROUP);
}


int H5UTLdataset_exists(hid_t loc_id, const char *name) {
  return _object_exists(loc_id, name, H5G_DATASET);
}


hid_t H5UTLopen_or_create_file(const char *file_name) {
  hid_t fid;
  if (file_exists(file_name)) {
    fid = H5Fopen(file_name, H5F_ACC_RDWR, H5P_DEFAULT);
  } 
  else {
    fid = H5Fcreate(file_name, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
  }
  return fid;
}


hid_t H5UTLopen_or_create_group(hid_t loc_id, const char *name) {
  hid_t gid_alns;
  if (!H5UTLgroup_exists(loc_id, name)) {
    gid_alns = H5Gcreate(loc_id, name, H5P_DEFAULT, H5P_DEFAULT, 0);
  } else {
    gid_alns = H5Gopen(loc_id, name, H5P_DEFAULT);
  }
  return gid_alns;
}


hid_t H5UTLcreate_1D_dataset(hid_t gid, const char *name, hid_t tid) {
  return H5UTLcreate_1D_dataset_size(gid, name, tid, 0);
}


hid_t H5UTLcreate_1D_dataset_size(hid_t gid, const char *name, hid_t tid, int init_size) {

  /* Some uses of this could be replaced with H5LTmake_dataset* functions */

  /* error if dataset exists */
  if (H5UTLdataset_exists(gid, name)) {
    return -1;
  }

  hsize_t dims[1] = { init_size };
  hsize_t maxdims[1] = { H5S_UNLIMITED };
  hid_t sid = H5Screate_simple(1, dims, maxdims);
  hid_t pid = H5Pcreate(H5P_DATASET_CREATE);
  hsize_t chunk_size[1] = { 1024 };
  H5Pset_chunk(pid, 1, chunk_size);
  H5Pset_deflate(pid, HDF_COMPRESSION_LEVEL);

  hid_t did = H5Dcreate(gid, name, tid, sid, H5P_DEFAULT, pid, H5P_DEFAULT);
  H5Pclose(pid);

  return did;
}


hid_t H5UTLcreate_2D_dataset(hid_t gid, const char *name, hid_t tid, int len) {
  return H5UTLcreate_2D_dataset_size(gid, name, tid, len, 0);
}


#define RANK_2D 2
hid_t H5UTLcreate_2D_dataset_size(hid_t gid, const char *name, hid_t tid, int len, int init_size) {

  /* Some uses of this could be replaced with H5LTmake_dataset* functions */

  /* error if dataset exists */
  if (H5UTLdataset_exists(gid, name)) {
    return -1;
  }

  hsize_t dims[RANK_2D] = { init_size, len };
  hsize_t maxdims[RANK_2D] = { H5S_UNLIMITED, len };
  hid_t sid = H5Screate_simple(RANK_2D, dims, maxdims);
  hid_t pid = H5Pcreate(H5P_DATASET_CREATE);
  hsize_t chunk_size[RANK_2D] = { HDF_CHUNK_SIZE, len };
  H5Pset_chunk(pid, RANK_2D, chunk_size);
  H5Pset_deflate(pid, HDF_COMPRESSION_LEVEL);
  hid_t did = H5Dcreate(gid, name, tid, sid, H5P_DEFAULT, pid, H5P_DEFAULT);
  H5Pclose(pid);

  return did;
}


int H5UTLget_1D_extent(hid_t sid) {
  /* could be replaced by H5LTget_dataset_ndims */
  hsize_t dims[1];
  H5Sget_simple_extent_dims(sid, dims, NULL);
  return dims[0];
}


void H5UTLget_2D_extent(hid_t sid, int *extent_0, int *extent_1) {
  /* could be replaced by H5LTget_dataset_ndims */
  hsize_t dims[2];
  H5Sget_simple_extent_dims(sid, dims, NULL);
  if (extent_0) *extent_0 = dims[0];
  if (extent_1) *extent_1 = dims[1];
}


void H5UTLwrite_1D_char_buf(hid_t did, const char* buf, int buf_num) {

  hid_t sid = H5Dget_space(did);
  hsize_t dims[1];
  H5Sget_simple_extent_dims(sid, dims, NULL);
  hsize_t slab_start[1] = { dims[0] };
  hsize_t slab_count[1] = { buf_num };

  dims[0] += buf_num;
  H5Dextend(did, dims);
  sid = H5Dget_space(did);
  
  H5Sselect_hyperslab(sid, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
  hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
  H5Dwrite(did, H5T_NATIVE_CHAR, mid, sid, H5P_DEFAULT, buf);
  H5Sclose(mid);
}


int H5UTLread_1D_char_buf(hid_t did, int pos, char* buf, int buf_num) {

  hid_t sid = H5Dget_space(did);
  hsize_t size = H5UTLget_1D_extent(sid);

  hsize_t slab_start[1] = { pos };
  int count = (((size - pos) < buf_num) ? (size - pos) : buf_num);

  if (count > 0) {
    hsize_t slab_count[1] = { count };
    H5Sselect_hyperslab(sid, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
    hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
    H5Dread(did, H5T_NATIVE_CHAR, mid, sid, H5P_DEFAULT, buf);
    H5Sclose(mid);
  }
  
  return count;
}


void H5UTLwrite_2D_char_buf(hid_t did, int num, const char* buf, int buf_num, int len) {

  hsize_t dims[2] = { num, len };
  H5Dextend(did, dims);
  hid_t sid = H5Dget_space(did);
  
  hsize_t slab_start[2] = { num - buf_num, 0 };
  hsize_t slab_count[2] = { buf_num, len };
  H5Sselect_hyperslab(sid, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
  hid_t mid = H5Screate_simple(2, slab_count, slab_count); 
  H5Dwrite(did, H5T_NATIVE_CHAR, mid, sid, H5P_DEFAULT, buf);
  H5Sclose(mid);
}


void H5UTLget_2D_char_len(hid_t did, int *str_len, int num) {

  hid_t sid = H5Dget_space(did);
  int str_num, max_len;
  H5UTLget_2D_extent(sid, &str_num, &max_len);

  int buf_size = HDF_RW_BUF_SIZE;
  char *buf = (char *)calloc(buf_size * max_len, sizeof(char));

  /* iterate over all sequences */
  char *str = (char *)calloc(max_len + 1, sizeof(char));
  int buf_offset = - buf_size;
  int i;
  for (i = 0; i < str_num; i++) {
    if (i >= buf_offset + buf_size) {
      /* fill buffer */
      buf_offset += buf_size;
      int buf_fill = str_num - buf_offset;
      if (buf_fill > buf_size) {
	buf_fill = buf_size;
      }
      hsize_t slab_start[2] = { buf_offset, 0 };
      hsize_t slab_count[2] = { buf_fill, max_len };
      H5Sselect_hyperslab(sid, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(2, slab_count, slab_count); 
      H5Dread(did, H5T_NATIVE_CHAR, mid, sid, H5P_DEFAULT, buf);
      H5Sclose(mid);
    }
    memcpy(str, buf + (i - buf_offset) * max_len, max_len);
    *(str + max_len) = 0;

    str_len[i] = strlen(str);
  }
  
}


/* TODO - replace this with H5LTset_attribute_int */
void H5UTLwrite_attr_int(hid_t did, const char *name, int value) {


#if 0
#endif
  hsize_t dim = 1;
  hid_t sid = H5Screate_simple(1, &dim, NULL);
  hid_t aid = H5Acreate(did, name, H5T_NATIVE_INT, sid, H5P_DEFAULT, H5P_DEFAULT);
  H5Awrite(aid, H5T_NATIVE_INT, &value);
  H5Aclose(aid);
  H5Sclose(sid);
}


void H5UTLsort_store_1D_dataset(hid_t did, hid_t tid, void *buf, int num, int typelen, int(*comparator)(const void *, const void *)) {

  /* sort the data */
  qsort(buf, num, typelen, comparator);

  /* write the buffer */
  hsize_t dims[1] = { num };
  H5Dextend(did, dims);
  H5Dwrite(did, tid, H5S_ALL, H5S_ALL, H5P_DEFAULT, buf);
}

